import pytest

from dataclasses import dataclass, field
from hologram import JsonSchemaMixin, ValidationError
from hologram.helpers import HyphenatedJsonSchemaMixin


@dataclass
class HasUnderscoreConverts(HyphenatedJsonSchemaMixin):
    a_thing: str = field(metadata={"preserve_underscore": True})
    other_thing: str


@dataclass
class ContainsHasUnderscoreConverts(JsonSchemaMixin):
    things: HasUnderscoreConverts


@pytest.fixture
def underscore():
    return HasUnderscoreConverts(a_thing="foo", other_thing="bar")


@pytest.fixture
def underscore_dict():
    return {"a_thing": "foo", "other-thing": "bar"}


@pytest.fixture
def contains(underscore):
    return ContainsHasUnderscoreConverts(things=underscore)


@pytest.fixture
def contains_dict(underscore_dict):
    return {"things": underscore_dict}


@pytest.fixture
def bad_dict():
    return {"a_thing": "foo", "other_thing": "bar"}


def test_base(underscore, underscore_dict, bad_dict):
    assert HasUnderscoreConverts.from_dict(underscore_dict) == underscore
    assert underscore.to_dict() == underscore_dict

    with pytest.raises(ValidationError):
        HasUnderscoreConverts.from_dict(bad_dict)


def test_nested(contains, contains_dict, bad_dict):
    assert ContainsHasUnderscoreConverts.from_dict(contains_dict) == contains
    assert contains.to_dict() == contains_dict

    with pytest.raises(ValidationError):
        ContainsHasUnderscoreConverts.from_dict({"things": bad_dict})



from dataclasses import fields
from enum import Enum
from typing import Type

from hologram import JsonSchemaMixin, FieldEncoder


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value

    # https://docs.python.org/3.6/library/enum.html#using-automatic-values
    def _generate_next_value_(name, start, count, last_values):
        return name


def StrLiteral(value: str) -> Type[StrEnum]:
    # mypy doesn't think this works, but it does
    return StrEnum(value, value)  # type: ignore


def register_pattern(base_type: Type, pattern: str) -> None:
    """base_type should be a typing.NewType that should always have the given
    regex pattern. That means that its underlying type ('__supertype__') had
    better be a str!
    """

    class PatternEncoder(FieldEncoder):
        @property
        def json_schema(self):
            return {"type": "string", "pattern": pattern}

    JsonSchemaMixin.register_field_encoders({base_type: PatternEncoder()})


class HyphenatedJsonSchemaMixin(JsonSchemaMixin):
    @classmethod
    def field_mapping(cls):
        result = {}
        for field in fields(cls):
            skip = field.metadata.get("preserve_underscore")
            if skip:
                continue

            if "_" in field.name:
                result[field.name] = field.name.replace("_", "-")
        return result


class ExtensibleJsonSchemaMixin(JsonSchemaMixin):
    ADDITIONAL_PROPERTIES = True



from dataclasses import dataclass
from typing import Tuple

from hologram import JsonSchemaMixin


@dataclass
class TupleMember(JsonSchemaMixin):
    a: int


@dataclass
class TupleEllipsisHolder(JsonSchemaMixin):
    member: Tuple[TupleMember, ...]


@dataclass
class TupleMemberFirstHolder(JsonSchemaMixin):
    member: Tuple[TupleMember, str]


@dataclass
class TupleMemberSecondHolder(JsonSchemaMixin):
    member: Tuple[str, TupleMember]


def test_ellipsis_tuples():
    dct = {"member": [{"a": 1}, {"a": 2}, {"a": 3}]}
    value = TupleEllipsisHolder(
        member=(TupleMember(1), TupleMember(2), TupleMember(3))
    )
    assert value.to_dict() == dct
    assert TupleEllipsisHolder.from_dict(dct) == value


def test_member_first_tuple():
    dct = {"member": [{"a": 1}, "a"]}
    value = TupleMemberFirstHolder(member=(TupleMember(1), "a"))
    TupleMemberFirstHolder.from_dict(dct) == value
    value.to_dict() == dct


def test_member_second_tuple():
    dct = {"member": ["a", {"a": 1}]}
    value = TupleMemberSecondHolder(member=("a", TupleMember(1)))
    TupleMemberSecondHolder.from_dict(dct) == value
    value.to_dict() == dct



import pytest

from dataclasses import dataclass
from typing import Union, Optional, List, Dict, Any
import re

from hologram import JsonSchemaMixin, ValidationError


@dataclass
class IHaveAnnoyingUnions(JsonSchemaMixin):
    my_field: Optional[Union[List[str], str]]


@dataclass
class IHaveAnnoyingUnionsReversed(JsonSchemaMixin):
    my_field: Optional[Union[str, List[str]]]


def test_union_decoding():
    for field_value in (None, [">=0.0.0"], ">=0.0.0"):
        obj = IHaveAnnoyingUnions(my_field=field_value)
        dct = {"my_field": field_value}
        decoded = IHaveAnnoyingUnions.from_dict(dct)
        assert decoded == obj
        assert obj.to_dict(omit_none=False) == dct

    # this is allowed, for backwards-compatibility reasons
    IHaveAnnoyingUnions(my_field=(">=0.0.0",)) == {"my_field": (">=0.0.0",)}


def test_union_decoding_ordering():
    for field_value in (None, [">=0.0.0"], ">=0.0.0"):
        obj = IHaveAnnoyingUnionsReversed(my_field=field_value)
        dct = {"my_field": field_value}
        decoded = IHaveAnnoyingUnionsReversed.from_dict(dct)
        assert decoded == obj
        assert obj.to_dict(omit_none=False) == dct

    # this is allowed, for backwards-compatibility reasons
    IHaveAnnoyingUnionsReversed(my_field=(">=0.0.0",)) == {
        "my_field": (">=0.0.0",)
    }


def test_union_decode_error():
    x = IHaveAnnoyingUnions(my_field={">=0.0.0"})
    with pytest.raises(ValidationError):
        x.to_dict(validate=True)

    with pytest.raises(ValidationError):
        IHaveAnnoyingUnions.from_dict({"my_field": {">=0.0.0"}})


@dataclass
class UnionMember(JsonSchemaMixin):
    a: int


@dataclass
class LongOptionalUnion(JsonSchemaMixin):
    # this devolves into Union[None, UnionMember]
    member: Optional[Union[None, UnionMember]]


def test_long_union_decoding():
    x = LongOptionalUnion(None)
    x.to_dict() == {"member": None}
    LongOptionalUnion.from_dict({"member": None})

    x = LongOptionalUnion(UnionMember(1))
    x.to_dict() == {"member": {"a": 1}}
    LongOptionalUnion.from_dict({"member": {"a": 1}}) == x

    with pytest.raises(ValidationError):
        try:
            LongOptionalUnion.from_dict({"member": {"b": 1}}, validate=False)
        except ValidationError as exc:
            str(exc)
            raise


@dataclass
class UnionDefinition(JsonSchemaMixin):
    my_field: Union[str, Dict[str, Any]]


def test_union_definition():
    dct = {"my_field": ["string_a", "string_b"]}
    with pytest.raises(ValidationError):
        try:
            UnionDefinition.from_dict(dct)
        except ValidationError as exc:
            assert exc.validator == "oneOf"
            assert re.search("'type': 'string'", str(exc))
            assert re.search("'type': 'object'", str(exc))
            raise



from typing import Optional, List, Dict, NewType, Any, Tuple, Union
from dataclasses import dataclass, field
from enum import Enum
from datetime import datetime
from uuid import UUID

from hologram import JsonSchemaMixin, FieldEncoder

Postcode = NewType("Postcode", str)


class PostcodeField(FieldEncoder):
    @property
    def json_schema(self):
        return {"type": "string", "minLength": 5, "maxLength": 8}


JsonSchemaMixin.register_field_encoders({Postcode: PostcodeField()})


class SubSchemas(JsonSchemaMixin):
    pass


class Weekday(Enum):
    MON = "Monday"
    TUE = "Tuesday"
    WED = "Wednesday"
    THU = "Thursday"
    FRI = "Friday"


@dataclass(eq=True)
class Point(SubSchemas):
    x: float = field(metadata=dict(description="Point x coordinate"))
    y: float = field(metadata=dict(description="Point y coordinate"))

    @classmethod
    def field_mapping(cls) -> Dict[str, str]:
        return {"x": "z"}


NewPoint = NewType("NewPoint", Point)


@dataclass
class Foo(SubSchemas):
    """A foo that foos"""

    a: datetime
    b: Optional[List[Point]]
    c: Dict[str, int]
    d: Weekday = field(init=False)
    f: Tuple[str, int]
    g: Tuple[str, ...]
    e: Optional[Postcode] = None
    h: Optional[NewPoint] = None


@dataclass(eq=True)
class Bar(JsonSchemaMixin):
    """Type with union field"""

    a: Union[Weekday, Point]


@dataclass
class Baz(JsonSchemaMixin):
    """Type with nested default value"""

    a: Point = field(default_factory=lambda: Point(0.0, 0.0))


@dataclass
class Recursive(JsonSchemaMixin):
    """A recursive data-structure"""

    a: str
    b: Optional["Recursive"] = None


@dataclass
class OpaqueData(JsonSchemaMixin):
    """Structure with unknown types"""

    a: List[Any]
    b: Dict[str, Any]


@dataclass
class Product(JsonSchemaMixin):
    name: str
    cost: float = 20.0


@dataclass
class ShoppingCart(JsonSchemaMixin):
    items: List[Product]

    @property
    def cost(self):
        return sum([item.cost for item in self.items])


@dataclass
class ProductList(JsonSchemaMixin):
    products: Dict[UUID, Product]


@dataclass
class Zoo(JsonSchemaMixin):
    """A zoo"""

    animal_types: Optional[Dict[str, str]] = field(default_factory=dict)



import pytest

from dataclasses import dataclass
from hologram import JsonSchemaMixin, ValidationError
from hologram.helpers import StrEnum, StrLiteral
from typing import List


class Bar(StrEnum):
    x = "x"
    y = "y"


@dataclass
class UsesBar(JsonSchemaMixin):
    bar: Bar


@dataclass
class UsesBarLiteral(JsonSchemaMixin):
    bar: StrLiteral(Bar.x)


def test_symmetry():
    def assert_symmetry(value):
        assert UsesBar.from_dict(value).to_dict() == value

    assert_symmetry({"bar": "x"})
    assert_symmetry({"bar": "y"})


def test_validation():
    with pytest.raises(ValidationError):
        UsesBar.from_dict({"bar": "invalid"})

    with pytest.raises(ValidationError):
        UsesBarLiteral.from_dict({"bar": "y"})


def test_schema():
    schema = UsesBar.json_schema()

    assert schema["type"] == "object"
    assert schema["required"] == ["bar"]
    assert schema["properties"] == {
        "bar": {"type": "string", "enum": ["x", "y"]}
    }

    schema = UsesBarLiteral.json_schema()

    assert schema["type"] == "object"
    assert schema["required"] == ["bar"]
    assert schema["properties"] == {"bar": {"type": "string", "enum": ["x"]}}
